# -*- coding: latin-1   -*-

#
# Questo modulo contiene classi e funzioni utili per organizzare test basati sulla gestione di messaggi
#

import sys, traceback
import types
import random
import operator

#
# Definizione del campo di un messaggio

# Un campo  definito da una stringa con uno dei seguenti formati:
# integer <dimensione in bytes>  <minimo> <massimo> 
# unsigned <dimensione in bytes>  <minimo> <massimo> ]
#    per gli interi le diomensioni possono essere 1,2,4
# float <dimensione in bytes>  <minimo> <massimo>
#    dimensione in byte pu essere 4 o 8 
# enum <dimensione in bytes> <valore> ... <valore>
# string <dimensione in bytes>
# array <max numero elementi> <tipo elemento> # dove il tipo elemento  uno degli altri tipi
# struct <nome struttura> dove il nome della struttura fa riferimento ad un altro messaggio
# union <nom campo> <nome struttura> dove <niome campo>  il nome del campo discriminante
#   e <nome struttura>  il nome della struttura-unione contenente un elemento per ogni valore
#   del discriminante, a partire da 0
# Minimi e massimi sono opzionali
# <max numero elementi> di un array pu essere il nome di un canmpo precedente, per array a dimensioni variabili (TBC)
#
#
class FieldDefinition(object):
    TYPES = ( "integer", "unsigned", "float", "enum", "string", "array", "struct", "union" )

    # Costruttore. Analizza e converte la stringa di definizione del campo
    def __init__(self, fname, description  ):
        self.values = []
        self.min = None; self.max = None
        self.element_type = None
        self.reference = None; self.refname = None
        self.disc_name = None; self.discriminant = None
        self.size = 0
        self.multiplicity = 1
        self.multiplicity_field = None
        
        try:
            self.name = fname
            desc_items = description.split()

            self.ftype = desc_items[0]
            assert self.ftype in self.TYPES , "Bad field type"

            if self.ftype == 'union':
                self.disc_name = desc_items[1].strip()
                sname = desc_items[2].strip()
                self.refname = sname

            elif self.ftype == 'struct':
                sname = desc_items[1].strip()
                self.refname =  sname
            else:

                try:
                    self.size = int(desc_items[1])
                except ValueError:
                    self.size = desc_items[1]
                
                if self.ftype in ('integer', 'unsigned'):
                    assert len(desc_items) <= 4, "Too many elements for  integer field"
                    assert self.size in (1,2,4,8), "Bad size for integer field"
                    if len(desc_items) > 2: self.min = int( desc_items[2] )
                    if len(desc_items) > 3: self.max = int( desc_items[3] )

                elif  self.ftype ==  "float":
                    assert len(desc_items) <= 4, "Too many elements for a floating point field"
                    assert self.size in (4,8), "Bad size for floating point field"
                    if len(desc_items) > 2: self.min = float( desc_items[2] )
                    if len(desc_items) > 3: self.max = float( desc_items[3] )
                
                elif self.ftype == "enum":
                    assert self.size in (1,2,4,8), "Bad size for enumerative field"
                    self.values = [ int(s) for s in desc_items[2:] ]

                elif self.ftype == "string":
                    pass
            
                elif self.ftype == "array":
                    assert len(desc_items) > 2, "Missing array type description"
                    self.element_type = FieldDefinition( self.name, " ".join( desc_items[2:] ) )
                    self.multiplicity = self.size
                    self.size = None # calcolata dinamicamente
                else:
                    raise AssertionError("Field type not supported")

        except AssertionError:
            raise Exception( "field %s : %s " % ( fname, sys.exc_info()[1] ) )
        except ValueError:
            raise Exception("field %s : bad format (%s)" % ( fname,description) )


    #
    # restituisce un valore random del tipo giusto per il campo
    #
    def get_random_value(self, default_array_size, values_dict = {} ):
        minval, maxval = self.min, self.max
        if self.ftype == 'integer':
            if minval == None: minval = -1*(2**self.size-1)
            if maxval == None: maxval = (2**self.size-1)
            return random.randint( minval, maxval )
        elif self.ftype == 'unsigned':
            if minval == None: minval = 0
            if maxval == None: maxval = (2**self.size-1)
            return random.randint( minval, maxval )
        elif self.ftype == 'enum':
            if minval == None: minval = 0
            if maxval == None: maxval = (2**self.size-1)            
            if self.values:
                return random.choice(self.values)
            else:
                return random.randint(minval, maxval)
            
        elif self.ftype == 'float':
            if not minval: minval = -10000.0
            if not maxval: maxval = +10000.0
            return ( minval + random.random()*(maxval-minval) )
        elif self.ftype == 'string':
            return ''.join( [chr(random.randint(0,255)) for _ in xrange(self.size)])        
        elif self.ftype == 'array':
            try:
                array_size = self.get_num_elements(values_dict)
            except ValueError:
                array_size = default_array_size
            return [ self.element_type.get_random_value(default_array_size, values_dict) for _ in xrange(array_size) ]
        elif self.ftype == 'struct':
            res = dict()
            for f in self.reference.fields:
                res[f.name] = f.get_random_value(default_array_size, res )
            return res
        elif self.ftype == 'union':
            try:
                disc_value = values_dict[self.disc_name]
            except KeyError:
                print "****", self.disc_name
                print "****", values_dict
                raise
            f = self.reference.fields[disc_value]
            res = dict()
            return  { f.name : f.get_random_value( default_array_size, values_dict ) } 
        else:
            return 0
        
     
    #
    # restituisce un valore  del tipo giusto per il metodo 'pack' con i campi settati come values_dict
    # oppure a zero se non presenti in values_dict
    #
    def partial_value(self,  value, values_dict  ):
        #print "****** partial_value: self.name = %s, value =  %s, values_dict = %s " % (self.name, value, values_dict)
        if self.ftype in ( 'integer', 'unsigned', 'enum'):
            if value: return value
            else: return 0
        elif self.ftype == 'float':
            if value: return value
            else: return 0.0
        elif self.ftype == 'string':
            if value: return value
            else: return ""
        elif self.ftype == 'array':
            try:
                array_size = self.get_num_elements(values_dict)
            except:
                array_size = 0
            if value: 
                return value
            else: 
                return [ self.element_type.partial_value(dict(), values_dict) for _ in xrange(array_size) ]
        elif self.ftype == 'struct':
            res = dict()
            for f in self.reference.fields:
                res[f.name] = f.partial_value(value.get(f.name, {}), res )
            return res
        elif self.ftype == 'union':
            try:
                disc_value = values_dict[self.disc_name]
            except KeyError:
                disc_value = 0

            f = self.reference.fields[disc_value]
            res = dict()
            return  { f.name : f.partial_value( value.get( f.name, {} ), values_dict ) } 
        else:
            return 0
           

    # Verifica che i riferimenti usati nella diefinizione di un campo corrispondano
    # a campi o strutture gia' definiti 
    def verify_reference(self, struct_dict, fields_dict = {} ):
        #print "************ verify_reference:", self.name, self.ftype
        if self.ftype == 'struct':
            self.reference = struct_dict.get( self.refname, None )
            if self.reference:
                pass
            else:
                raise Exception("Structure %s referenced in message %s not found" % ( self.refname, self.name ) )
            
        elif self.ftype == 'union':
            self.discriminant = fields_dict.get( self.disc_name, None )
            if not self.discriminant:
                raise Exception("Field %s referenced in message %s not found" % ( self.discname, self.name ) )
            
            self.reference = struct_dict.get( self.refname, None )
            if not self.reference:
                raise Exception("Union %s referenced in message %s not found" % ( self.refname, self.name ) )
            
            # imposta automaticamente i limiti del discriminante
            exp_min = 0
            if self.discriminant.min == None:
                self.discriminant.min = exp_min
            else:
                assert self.discriminant.min == exp_min, "The minimum value of discriminant field %s shall be %d" % (self.disc_name, exp_min)
            
            exp_max = len(self.reference.fields) -1
            if self.discriminant.max == None:
                self.discriminant.max = exp_max
            else:
                assert self.discriminant.max == exp_max, "The maximum value of discriminant field %s shall be %d" % (self.disc_name, exp_max) 
            
            
            # verifica che tutte le alternative di una union abbiano la stessa dimensione
            values_dict = {}
            sizes_dict = dict( (f.name, f.compute_size(values_dict)) for f in self.reference.fields )
            max_size = max ( sizes_dict.values() )
            for name, size in sizes_dict.items():
                if size != max_size:
                    raise Exception( "In union %s field %s has size %s instead of %s" % ( self.name, name, size, max_size ) )

        elif self.ftype == 'array':
            if self.element_type.ftype in ( 'struct', 'union'):
                    self.element_type.verify_reference(struct_dict)
            try:
                _ = int(self.multiplicity)
            except:
                try:
                    self.multiplicity_field = fields_dict[ self.multiplicity ]
                except KeyError:
                    raise Exception("The size of array field %s is not a field name (%s)" %( self.name, self.multiplicity) )


    #restituisce il numero di elementi di un campo
    def get_num_elements(self, values_dict ):
        if self.ftype == 'array':
            try:
                res = int(self.multiplicity)
            except ValueError:
                field_value=None
                try:
                    field_value = values_dict[self.multiplicity]
                    res = int( field_value )
                except (ValueError, TypeError):
                    raise Exception("The value of field %s (%s) cannot be used as dimension of array field %s" % (self.multiplicity,
                                                                                                                  str(field_value),
                                                                                                                  self.name))
                except KeyError:
                    raise Exception("The size of array field %s is not a field name (%s)" %( self.name, self.multiplicity) )                
        else:
            res = 1
            
        return res
                
                
    # Verifica se un dizionario ha i campi di riferimento necessari ( dimensione dell'array e valore dei discrminante) 
    # Se nn li ha li aggiunge
    def find_add_reference_fields(self, values_dict):
        if self.ftype == 'union':
            disc_value = values_dict.get( self.disc_name, None )
            if disc_value == None : values_dict[self.disc_name] = 0
        elif self.ftype == 'array':
            try:
                _ = self.get_num_elements(values_dict)
            except:
                values_dict[self.multiplicity] = self.multiplicity_field.max
            
                

    # restituisce il formato con cui leggere/scrivere il campo da/in un pacchetto di bytes
    # Il parametro parent_values_dic contiene l'insieme dei campi della struttura superiore;
    # questo parametro e' necessario in caso di campi union e di array a dimensione variabile  
    def struct_format(self, parent_values_dict  ):    
        self.find_add_reference_fields(parent_values_dict)                
        res = "?"
        if self.ftype in ("integer", "unsigned", "enum"):
            if self.size == 1: res = "b"
            elif self.size == 2: res = "h"
            elif self.size == 4: res = "i"
            elif self.size == 8: res = "l"
            else: raise Exception("Unsupported integer/unsigned/enum size: %s" % str(self.size) )
            if self.ftype == "unsigned": res = res.upper()
        elif self.ftype == "float":
            if self.size == 4 : res = "f"
            elif self.size == 8: res = "d"
            else: raise Exception("Unsupported float size: %s" % str(self.size) )
        elif self.ftype == "string":
            res = "%ds" % self.size
        elif self.ftype == "array":
            num_elements = self.get_num_elements(parent_values_dict)
            try:
                field_value = parent_values_dict[self.name]
            except KeyError:
                field_value = [dict() for _ in range(num_elements)]
            
            res = ""
            assert len(field_value) >= num_elements, "Not enough values for field %s. Expected %d, got %d" % ( self.name, num_elements, len(field_value) ) 
            for idx in range(num_elements):
                elem_value = { self.name : field_value[idx] }
                res += self.element_type.struct_format(elem_value)
            #res = num_elements * self.element_type.struct_format(values_dict) # array of array not supported
                
                
        elif self.ftype == "struct":
            assert self.reference, "Referenced structure %s not found for field %s" % ( self.refname, self.name )             

            try:
                field_value = parent_values_dict[self.name]
            except KeyError:
                field_value = dict() 
            
            field_formats = []
            for f in self.reference.fields:
                field_formats.append(f.struct_format(field_value  ) )
            res = "".join( field_formats )           
            #disc_value = struct_values.get( self.disc_name, None )
            #res = "".join( [f.struct_format(struct_values, disc_value) for f in self.reference.fields] )
            
        elif self.ftype == 'union':
            assert self.reference, "Referenced structure %s not found for field %s" % ( self.refname, self.name )
            try:
                disc_value = parent_values_dict[self.disc_name]
            except KeyError:
                raise Exception("Value of discriminant field %s not specified when computing format of field %s" %( self.disc_name, self.name ))
            assert 0 <= disc_value < len(self.reference.fields),  "Bad value for discriminant field %s of union %s : %s" % ( 
                                                                                            self.disc_name, self.name, str(disc_value) )
            selected_field =  self.reference.fields[disc_value]
            field_value =  parent_values_dict.get( self.name, dict() )
            selected_field_value = field_value.get( selected_field.name, dict() )       
            res = selected_field.struct_format(field_value)
            
        else:
            raise Exception("Unknown structure for field %s of type %s" % ( self.name, self.ftype ))
        return res


    #
    # Restituisce la dimensione del campo
    #
    def compute_size(self, values_dict = None  ):

        if not values_dict: values_dict = {}

        if self.ftype == 'union': # il campo  una union
            #fmt_list = [ '!'+ f.struct_format(values_dict, idx) for (idx, f) in enumerate(self.reference.fields) ]
            fmt_size = max( f.compute_size(values_dict) for f in self.reference.fields )
        elif self.ftype == 'struct':
            #Se per un campo deklla struttura non  specificato alcun valore, si specifica il massimo
            # Questo serve per poter dimensionare i campi array a dimensione variabile
            for f in self.reference.fields:
                if f.max != None:
                    if f.name not in values_dict.keys():
                        values_dict[f.name] = f.max         
            
            fmt_size = sum ( f.compute_size(values_dict) for f in self.reference.fields )
        else:
            fmt =    '!'+ self.struct_format(values_dict)
            fmt_size =  struct.calcsize( fmt )
        return fmt_size


    # determina quanti valori vanno presi da una lista ottenuta spacchettando un messaggio
    # restituisce il valore del campo e la lista dei valori residui
    # Usata nella versione iniziale di unpack, ora non piu' usata ma potrebbe servire
    def assign_values(self, values, num_elements = None, values_dict = {}, disc_value = None ):
        #print "*******", self.name, self.ftype, values
        if self.ftype in("integer", "unsigned", "enum", "float"):
            return values[0], values[1:]
        elif self.ftype == "string":
            return values[0], values[1:] 
            #return "".join( values[:self.size]  ), values[self.size:]
        elif self.ftype == "array":
            assert num_elements != None, "Unknown mumber of elements for array field %s" % self.name
            res = []; vlist = values[:]
            for n in xrange(num_elements):
                v, vlist =  self.element_type.assign_values(vlist)
                res.append(v)
            return res, vlist
        elif self.ftype == 'struct':
            res = dict()
            for f in self.reference.fields:
                disc_value = res.get( f.disc_name, None )    
                res[f.name], values = f.assign_values( values, f.get_num_elements( dict()), values_dict, disc_value )
            return res, values
        elif self.ftype == 'union':
            #print "****", values_dict
            assert disc_value != None, "No value provided for discriminant field %s" % self.disc_name
            assert (0 <= disc_value < len(self.reference.fields )), "Bad value for dicriminant value:"+ str(disc_value)
            res = dict()
            f = self.reference.fields[disc_value]
            res[f.name], values = f.assign_values( values, f.get_num_elements( dict()), values_dict, disc_value )
            return res, values
        else:
            raise Exception("assign_values - unsupported type: %s", self.ftype )
                

    # Destruttura il valore i un campo, restituendo la lista dei valori dei suoi elementi base
    # ( operazione inversa al metodo 'assign_values' )
    def flatten_value( self, value, num_elements = None, values_dict= {}, disc_value = None ):
        if self.ftype in("integer", "unsigned", "enum", "float"):
            return [value,]
        elif self.ftype == "array":
            assert num_elements != None, "Unknown mumber of elements for array field %s" % self.name
            res = []
            for n in xrange(num_elements):
                res.extend( self.element_type.flatten_value( value[n] ) )
            return res
        elif self.ftype == "string":
            return [value,]
        elif self.ftype == "struct":
            values = []
            for f in self.reference.fields:
                try:
                    field_num_elements = f.get_num_elements(value) # not sure this work for inner variable size array
                    disc_value = value.get(f.disc_name, None) 
                    values.extend( f.flatten_value( value[f.name], field_num_elements, values_dict, disc_value  ) )
                except KeyError:
                    raise Exception( "Missing value for field %s" % f.name  ) 
            return values            
        elif self.ftype == "union":
            assert self.reference != None, "No reference field for union"
            assert disc_value != None, "Could not find a value for discriminant field %s" % self.disc_name
            assert (0 <= disc_value < len(self.reference.fields )), "Bad value for dicriminant value:"+ str(disc_value)
            f = self.reference.fields[disc_value]
            try:
                field_num_elements = f.get_num_elements(dict())
                values = f.flatten_value( value[f.name], field_num_elements, values_dict  )
            except KeyError:
                raise Exception( "Missing value for field %s" % f.name  )
            return values            
        else:
            raise Exception("flatten_value - unsupported type: %s", self.ftype )

    #
    # Trasforma un dizionario di campi-valiori in una sequenza di bytes
    #
    def pack(self, values_dict, byte_order="!" ):
        disc_value = values_dict.get( self.disc_name, None )
        num_elements = self.get_num_elements(values_dict)           
        fmt = self.struct_format(values_dict)
        try:
            value = values_dict[self.name]
            flatten_value = self.flatten_value(value, num_elements,  values_dict, disc_value )
            return struct.pack( byte_order+fmt, *flatten_value )
        except KeyError:
            raise Exception("No value provided for field name %s" % self.name )
        except:
            err, det, tb = sys.exc_info()
            print "********", err, det
            traceback.print_tb(tb)
            #print '******** structuded value', value
            #print "******** flattened value size", len(flatten_value)
            #print "******** format string", fmt
            raise Exception("Error packing field %s: %s" % (self.name, det) )

    #
    # Trasforma un sequenza di byte in un valore o un dizionario campi-valori 
    #
    def unpack(self, bytes, values_dict, byte_order="!" ):
        used_bytes = 0
        try:
            if ( self.ftype == 'struct' ):
                used_bytes = 0
                value = dict() 
                for f in self.reference.fields:
                    f_value, f_used_bytes = f.unpack( bytes[used_bytes:], value, byte_order)
                    value[f.name] = f_value
                    used_bytes += f_used_bytes 
            elif ( self.ftype == 'union' ):
                disc_value = values_dict.get( self.disc_name, None )
                if disc_value == None: print "****", self.name, self.disc_name, values_dict
                assert disc_value != None, "Could not find a value for discriminant field %s" % self.disc_name
                assert (0 <= disc_value < len(self.reference.fields )), "Bad value for dicriminant value:"+ str(disc_value)
                f = self.reference.fields[disc_value]
                u_value, used_bytes =  f.unpack( bytes[used_bytes:], values_dict, byte_order)
                value= { f.name : u_value }
            elif self.ftype == 'array' :
                    num_elements = self.get_num_elements( values_dict )
                    value = []
                    fmt_size = self.element_type.compute_size(values_dict)
                    for i in range(num_elements):
                        v, used =  self.element_type.unpack(bytes[used_bytes:], values_dict, byte_order )
                        value.append(v)
                        used_bytes += used              
            else:
                fmt = self.struct_format(values_dict )
                fmt = byte_order+fmt
                used_bytes = struct.calcsize(fmt)
                value = struct.unpack( fmt, bytes[:used_bytes] )
                value = value[0]
                                
            return value, used_bytes
        
        except:
            err, det, tb = sys.exc_info()
            print "*********", err, det
            traceback.print_tb(tb)
            #print '******** bytes', bytes[:fmtsize]
            #print "******** format string", fmt
            print "******** len(bytes) = ",len(bytes)
            print "******** used_bytes = ", used_bytes      
            #print "******** flattened value", value
            #print "******** structured value", struct_value
            raise Exception("Error unpacking field %s: %s" % (self.name, det) )

                            
    # Restituisce una rappresentazione del campo equivalente a quella letta dal file
    def __str__(self):
        if self.ftype == "struct":
            res = "%s = %s %s" % ( self.name, self.ftype, self.refname )
        elif self.ftype == "union":
            res = "%s = %s %s %s" % ( self.name, self.ftype, self.disc_name, self.refname )
        else:
            size = self.size
            if self.ftype == "array" : size = self.multiplicity
            res = "%s = %s %s" % ( self.name, self.ftype, str(size) )
            if self.min != None : res += " %s" % str(self.min)
            if self.max != None: res += " %s" % str(self.max)
            if self.values: res += " "+ " ".join(str(v) for v in self.values)
            if self.element_type: res += " " + str(self.element_type).split("=")[1]
        return res
        



#
# Definizione del messaggio o di una struttura
#
class MessageDefinition(object):
    # Crea un messaggio senza campi
    def __init__(self, mtype, mname ):
        self.mtype = mtype
        self.name = mname
        self.fields = []
        self.spare_counter = 1
        self.byte_order_char = "!" # network order unless differentrly specified

    def set_byte_order(self, byte_order_char):
        self.byte_order_char = byte_order_char

    # Aggiunge un campo alla definizione del messaggiio
    def define_field( self, fname, description  ):
        if fname == "spare":
            fname = "spare_%d" % self.spare_counter
            self.spare_counter += 1
        f = FieldDefinition( fname, description   )
        self.fields.append( f )


    # Verifica i campi con riferimenti ad altre struttura
    def verify_references(self, struct_dict):
        fields_dict = dict( (f.name, f) for f in self.fields )
        for f in self.fields:
            fsize = f.verify_reference(struct_dict, fields_dict )
            #print "***** MessageDefinition.verify_reference ***", f.name, fsize
            #f.verify_reference(struct_dict)
                                           

    # Generatore sui nomi dei campi
    def iterate_field_names(self):
        for f in self.fields:
            yield f.name


    # Estrae da una stringa di bytes i campi del messaggio
    # Restituisce un dizionario {nome_campo:valore_campo}
    def unpack( self, bytes ):
        field_values = dict()
        for f in self.fields:
            assert len(bytes) != 0, "Cannot unpack field %s: no more bytes" % f.name
            value, used_bytes  = f.unpack( bytes, field_values, self.byte_order_char )            
            bytes = bytes[used_bytes:]
            field_values[f.name] = value
            #print '************* %s = %s' % (f.name, value)
        assert len(bytes) == 0, "Unused %d bytes in unpacking the message" % len(bytes)
        return field_values
    
    # Comverte una serie di valori in una stringa di bytes
    # I valori in input sono dati come nomme_campo=valore_campo
    # Sottostrutture sono dizionari {name_campo:valore_campo}, array sono liste
    def pack( self, **field_values  ):
        bytes = []
        for f in self.fields:
            bytes += f.pack( field_values, self.byte_order_char )
        return "".join( bytes )


    # Legge un file di configurazione e restituisce la lista di messaggi in esso definita
    # Nota : ConfigParser non si pu usare perch non mantiene l'ordine dei campi
    @classmethod
    def read_config( self, fname ):
        struct_dict = dict()
        mlist = []
        in_f = file( fname, "r" )
        msg = None
        line_cnt = 1
        for line in in_f:
            try:
                line = line.split("#")[0].strip()
                if not line: continue
                if '[' in line and ']' in line: # start section
                    sect_name = line.split("[")[1].split("]")[0].strip()
                    mtype, mname = [ e.strip() for e in sect_name.split() ]
                    if msg:
                        mlist.append (msg)
                        struct_dict [ msg.name ] = msg
                    msg = MessageDefinition(mtype, mname)
                elif '=' in line: # field line
                    fname, fdesc = [s.strip() for s in line.split("=")]
                    msg.define_field( fname, fdesc )
                else:
                    raise Exception("Bad Line Format %s" %  line )
            except ValueError:
                raise Exception("Line %d : Bad Line Format %s" %( line_cnt, line) )
            except:
                err, det, tb = sys.exc_info()
                traceback.print_tb(tb)
                raise Exception("Line %s : %s" % (line_cnt, det ))
            
            
            line_cnt +=1
        if msg:
            mlist.append( msg )
            struct_dict[msg.name] = msg
        for m in mlist: m.verify_references( struct_dict )
        return mlist
       
    
    # Restituisce la definizione di uno specifico campo
    def get_field_definition(self, field_name):
        res = None
        for f in self.fields:
            if f.name == field_name:
                res = f; break
        return res
        
    
    # Restituisce una rappresentazione del messaggio molto simile a quella letta dal file di configurazione
    def __str__(self):
        res = ["[%s %s]" % (self.mtype, self.name)]
        for f in self.fields:
            res.append("\t" + str(f))
        return "\n".join(res)
    


    #
    # Calcola il numero di byte corrispondenti al messaggio
    # In caso di un messaggio/struttura a dimensione variable, occorre specificare
    # il parametro disc_fields, che deve contenere i valori dei campi che 
    # determinano la dimensione
    #
    def compute_size(self, disc_fields = dict() ):
        size_list = [ f.compute_size(disc_fields) for f in self.fields ]
        #print "******",  size_list
        return sum(size_list)
     
    #
    # Calcola un insieme di valori random per il messaggio
    #
    def get_random_value(self, max_array_size):
        res = dict( (f.name, f.get_random_value(max_array_size)) for f in self.fields ) 
        return res

    # Restituisce un insieme di valori per il messaggio, parzialmente settati come nel dizionario di input
    def partial_value(self, values_dict):
        res = dict()
        for f in self.fields:
            res[f.name] = f.partial_value( values_dict.get(f.name, {}), res  )
        return res




#
# Stampa una struttura di dizionari nidificato
def print_field_value( name, value, tab_level='' ):
    if type(value) != types.DictionaryType:
        print "%s%s = %s" % ( tab_level, name, value )
    else:
        print "%s= " %(name)
        for  n, v in value.items():
            print_field_value( name+"."+n, v, tab_level+"   ") 
   


        
#
# ------- Test code after this line ----------------------------------------
#

import sys, os, os.path
import random, struct

UWP_ROOT = os.environ.get("UWP_ROOT", os.path.realpath( os.path.join("..")) )


def almost_equal(x,y):
    if type(x) == types.FloatType:
        return abs(x-y) < 0.001
    elif type(x) == types.ListType:
        return reduce( operator.and_, [almost_equal(x[i], y[i]) for i in range(len(x)) ] )
    elif type(x) == types.DictType():
        return reduce( operator.and_, [ almost_equal(x[n], y[n]) for n in x.keys() ] )
    else:
        return x == y


def test_message_def( do_print=0):
    fname =  "../test/test_message_def.txt"
    message_list = MessageDefinition.read_config( fname )
    for m in message_list:
        print "Read Message %s (size %d) " % ( m.name, m.compute_size() )
        if do_print: print m
    return message_list



def test_pack_unpack():
    message_list = test_message_def()
    for msg in message_list:
        if msg.mtype == 'UNIONE': continue
        print "test_pack_unpack", msg.name
        try:
            in_values = dict()
            for f in msg.fields:
                in_values[f.name] = f.get_random_value(100, in_values)
            bytes = msg.pack( **in_values )
            out_values = msg.unpack(bytes)
            for in_name, in_value in in_values.items():
                out_value = out_values[in_name]
                if type(out_value) == types.FloatType:
                    err_str = "Message %s: Value mismatch in field %s (\n\texpected:%s \n\tfound: %s)" % (
                        msg.name,
                        in_name,
                        str(in_value),
                        str(out_value)
                        )
                    assert almost_equal(in_value, out_values[in_name]), err_str
        except:
            print "Errore in packing/unpacking la struttura %s" % msg.name
            err, det,  tb = sys.exc_info()
            print err, det
            traceback.print_tb(tb) 
    

def test():
    test_pack_unpack()


if __name__ == '__main__':
    test()

