#
# Questo modulo effettua il parsing di un file di tipo '.ptl' o '.mdl'
# (il formato dei file contenti i modelli di RatiionalRose ) 
# e ne mette a disposizione il contenuto sotto forma di un oggetto ParseTree.
# Un parse tree e' composto da una struttura nidificata, in cui ogni elemento e' una tupla (...). 
# Il modulo fornisce anche una classe PTLObject, che puo' essere usata per descrivere gli elementi del
# parse tree del tipo (object ...).
# 
#

import sys, traceback, types
from  ply import yacc, lex


#
# Lexical tokens
#

KEYWORDS = ('OBJECT', 'LIST')
KEYWORDS_MAP = dict( (x.lower(), x ) for x in KEYWORDS )

tokens = KEYWORDS + ( "LPAREN", "RPAREN",  "QUOTED", "DOCLINE", "DOTNUMBER", "REFERENCE", "IDENT", "COMMA" )
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_DOTNUMBER = r'([+|-]*[0-9]|\.)+'
t_REFERENCE = r'@[0-9]+'
t_COMMA = r'\,'



def t_QUOTED(t):
    r'".*"'
    t.value = t.value[1:-1]
    return t

def t_DOCLINE(t):
    r'\|.*'
    t.value = t.value[1:]
    return t

def t_IDENT(t): 
    '[A-Za-z_][\w_]*'
    t.type = KEYWORDS_MAP.get(t.value, "IDENT")
    return t

t_ignore = ' \t\r'

def t_error(t):
    print 'Lexical error:', str(t)[:300]

def t_NEWLINE(t):
    r'\n+'
    t.lexer.lineno += len(t.value)

def t_TWOLINEQUOTED(t):
    r'".*\n.*"'

#
# Grammar rules
#

_TRACE=0
def TRACE(*args):
    if _TRACE:
        for a in args: print str(a)[:300],
        print

def p_list_sequence(p):
    """
    list_sequence : list_sequence list
                  | list
    """
    if len(p) == 2:
        p[0] = [p[1],]
    else:
        p[0] = p[1]; p[0].append(p[2])


def p_list_as_object(p):
    """ 
    list : LPAREN OBJECT IDENT QUOTED REFERENCE field_list RPAREN
    """
    TRACE( 'p_list_as_object', p[3] )
    p[0] = (p[2], p[3], p[4], p[5], p[6])


def p_list_as_object_object_noreference(p):
    """ 
    list : LPAREN OBJECT IDENT QUOTED field_list RPAREN
    """
    TRACE( 'p_list_as_object_noreference', p[3] )
    p[0] = (p[2], p[3], p[4], None, p[5])

def p_list_as_object_noname(p):
    """ 
    list : LPAREN OBJECT IDENT REFERENCE field_list RPAREN
    """
    TRACE( 'p_list_as_object_noname', p[3] )
    p[0] = (p[2], p[3], None, p[4], p[5])

def p_list_as_object_noname_noreference(p):
    """ 
    list : LPAREN OBJECT IDENT field_list RPAREN
    """
    TRACE( 'p_list_as_object_noname_noreference', p[3] )
    p[0] = (p[2], p[3], None, None, p[4])


def p_list_as_object_nofields(p):
    """ 
    list : LPAREN OBJECT IDENT QUOTED REFERENCE RPAREN
    """
    TRACE( 'p_list_as_object_nofields', p[3] )
    p[0] = (p[2], p[3], p[4], p[5], [] )


def p_list_as_object_nofields_noreference(p):
    """ 
    list : LPAREN OBJECT IDENT QUOTED RPAREN
    """
    TRACE( 'p_list_as_object_nofields_noreference', p[3] )
    p[0] = (p[2], p[3], p[4], None, [] )



def p_list_as_list(p):
    """
    list : LPAREN LIST IDENT RPAREN
         | LPAREN LIST IDENT list_sequence RPAREN
    """
    TRACE( 'p_list_as_list', p[3] )
    if len(p) == 5:
        p[0] = (p[2], p[3])
    else:
        p[0] = (p[2], p[3], p[4])


def p_list_as_list_noname(p):
    """
    list : LPAREN LIST list_sequence RPAREN
    """
    p[0] = ( p[2], p[3] )


def p_list_as_list_of_values(p):
    """
    list : LPAREN value_sequence RPAREN
    """
    TRACE( 'p_list_as_list_of_values' )
    p[0] = p[2]


def p_list_as_coordinates(p):
    """
    list : LPAREN comma_value_sequence RPAREN
    """
    

def p_value_sequence(p):
    """
    value_sequence : value_sequence field_value
                   | field_value
    """
    if len(p) == 2:
        p[0] = [p[1],]
    else:
        p[0] = p[1]; p[0].append(p[2])


def p_comma_value_sequence(p):
    """
    comma_value_sequence : value_sequence COMMA field_value
                         | field_value
    """
    if len(p) == 2:
        p[0] = [p[1],]
    else:
        p[0] = p[1]; p[0].append(p[2])




def p_field_list(p):
    """
    field_list : field
               | field_list field
    """
    if len(p) == 2:
        p[0] = [p[1],]
    else:
        p[0] = p[1]; p[1].append(p[2])

def p_field(p):
    """
    field : IDENT field_value
    """
    TRACE( 'p_field:', p[1], str(p[2])[:80] )
    p[0] = (p[1], p[2])


def p_field_value(p):
    """
    field_value : QUOTED
                | DOTNUMBER
                | REFERENCE
                | IDENT 
                | list
                | documentation
    """
    TRACE( 'field_value:', p[1] )
    p[0] = p[1]



def p_documentation(p):
    """
    documentation : documentation DOCLINE
                  |  DOCLINE
    """
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = p[1]+'\n'+p[2]


def p_error(p):
    print 'Parser error:', p
    return p



#
# Oggetto generico definito in un file 'PTL'
#

class PTLObject(object):
    def __init__(self, node ):
        assert node[0] == 'object', "PTLObject.__init__ : invalid node"
        _, self.otype, self.oname, self.oref, el_list = node
        self.elements = dict()
        for aname, avalue in el_list:
            if avalue and (avalue[0] == 'list'): avalue = list(avalue[1:])
            self.elements[aname] = avalue
        
    def get_attribute(self, attr_name):
        try:
            return self.attributes[attr_name]
        except KeyError:
            return None
        
    def __repr__(self):
        res = '(object '+  ' '.join( [ str(x) for x in (self.otype, self.oname, self.oref) ]) + '\n'
        for k, v in self.elements.items():
            res += k + ' = ' + str(v) + '\n'
        return res
        

#
# Parser tree class
#
class ParseTree(object):
    def __init__(self, root=None ):
        self.root = root

    def parse_file(self, file_name):
        l = lex.lex()
        p = yacc.yacc(start='list_sequence')
        f = open(file_name)
        data = ''.join( f.readlines() )
        f.close()
        self.root = p.parse( data )        

    def traverse(self, matching_tuple = None):
        to_do = list( self.root )
        while to_do:
            n = to_do.pop(0)
            #try: 
            #    if n[0] == 'object': print str(n)[:200]
            #except TypeError: 
            #    pass
            if not matching_tuple:
                found = True
            else:
                # Compare the node with the matching tuple
                found = True
                try:
                    list_n = list(n)
                    if len(list_n) < len(matching_tuple): 
                        found = False
                    else:
                        for e1, e2 in zip(list_n, matching_tuple):
                            if (e2 != None) and (e1 != e2):
                                found=False; break
                except TypeError:
                    found = False

            if found: yield n

            if type(n) in (types.TupleType, types.ListType):
                to_do+= list(n)

    def get_package(self, full_name):
        tree = self
        TRACE( 'get_package:', 'full_name = ', full_name )
        name_list = full_name.split('::')
        # For some reason, Logical View is an object of type Design
        if name_list[0] == 'Logical View':
            otype = 'Design' 
        else:
            otype =  'Class_Category'
        while name_list:
            name = name_list.pop(0)
            found=False
            TRACE( 'get_package:', 'Searching object', name, 'of type', otype  )
            for node in tree.traverse( ('object', otype) ):
                TRACE( 'get_package:', 'Comparing with', node[2] )
                #print '***', node[2]
                if name == node[2]:
                    TRACE( 'get_package:', 'Found ', name )
                    found = True; break
            if found == False: break
            else: 
                otype = 'Class_Category'
                tree = ParseTree(node)

        if found:
            return ParseTree(node)
        else:
            return None
                                

    def get_package_elements( self, package_name, matching_tuple=None ):
        p = self.get_package( package_name )
        if p == None:
            raise Exception( "Cpuld not find category " + package_name)
        for e in p.traverse( matching_tuple ):
            yield e
                                
                    



#
# ---- Test code after this line -----------------------------------

TEST_LEX_DATA = """
(object Petal
    version    	47
    _written   	"Rose 8.0.0303.1400"
    client      @12
    documentation 
|         esempio di stringa di documentazione
|
    charSet    	0)

"""
def test_lexical():
    l = lex.lex()
    l.input( TEST_LEX_DATA)
    while 1:
        t = l.token()
        print t
        if not t: break




PARSER_TEST_DATA = """


(object Petal
    version    	47
    _written   	"Rose 8.0.0303.1400"
    pippo       32.1
    value       (1 2 3)
    value       (value Text "")
    charSet    	0)

(object Design "Logical View"
    attributes 	(list Attribute_Set
	(object Attribute
	    tool       	"Java"
	    name       	"IDE"
	    value      	"Internal Editor")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagName1"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagText1"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagApply1"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagName2"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagText2"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagApply2"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagName3"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagText3"
	    value      	"")
	(object Attribute
	    tool       	"Java"
	    name       	"UserDefineTagApply3"
	    value      	""))
    quid       	"36D487460258"
    defaults   	(object defaults
	rightMargin 	0.250000
	leftMargin 	0.250000
	topMargin  	0.250000
	bottomMargin 	0.500000
	pageOverlap 	0.250000
	clipIconLabels 	TRUE
	autoResize 	TRUE
	snapToGrid 	TRUE
	gridX      	16
	gridY      	16
	defaultFont 	(object Font
	    size       	9
	    face       	"Arial"
	    bold       	FALSE
	    italics    	FALSE
	    underline  	FALSE
	    strike     	FALSE
	    color      	0
	    default_color 	TRUE)
	showMessageNum 	1
	showClassOfObject 	TRUE
	notation   	"Unified")
    root_usecase_package 	(object Class_Category "Use Case View"
	quid       	"34DBB4830142"
	documentation 	
|The Use Case View organizes the Use-Case Model and the optional Business Use-Case Model.
|

))

"""


def test_parser():
    l = lex.lex()
    p = yacc.yacc(start='list_sequence')
    res = p.parse( PARSER_TEST_DATA, debug=2 )
    print 'Parse result:'; print res


TEST_PARSER_FILE = "../test/smf.mdl"


def test_lexer_with_file():
    l = lex.lex()
    f = open(TEST_PARSER_FILE)
    #data = merge_continuation_lines(f.readlines() )
    data = ''.join(f.readlines())
    f.close()
    l.input(data)
    while 1:
        t = l.token()
        if t: print t
        else: break
    


def test_parser_with_file():
    global _TRACE
    l = lex.lex()
    p = yacc.yacc(start='list_sequence')
    f = open(TEST_PARSER_FILE)
    data = ''.join( f.readlines() )
    f.close()
    _TRACE=1
    res = p.parse( data )
    _TRACE=0


def test_parse_tree():
    m = ParseTree(); 
    m.parse_file(TEST_PARSER_FILE)
    for e in m.traverse( ('object', 'UseCase')  ):
        print 'Found use case:', str(e)[:300]
    for e in m.traverse( ('object', 'Class')  ):
        print 'Found class:', str(e)[:300]
    

def test_traverse_categories():
    m = ParseTree(); 
    m.parse_file(TEST_PARSER_FILE)
    for e in m.traverse( ('object', 'Class_Category') ):
        print e[:3]


def test_get_package():
    global _TRACE
    m = ParseTree(); 
    m.parse_file(TEST_PARSER_FILE)
    _TRACE=0
    pname = "Use Case View::Capabilities"
    p = m.get_package( pname)
    assert p != None, "Package %s not found" % pname
    pname = "Logical View::Design Model"
    p = m.get_package( pname)
    assert p != None, "Package %s not found" % pname


def test_get_package_classes():
    m = ParseTree(); 
    m.parse_file(TEST_PARSER_FILE)
    for c in m.get_package_elements( "Use Case View::SMF-Actors", ("object", "Class") ):
        o = PTLObject(c)
        print o
        
    

def test():
    test_suite = (
#        test_lexical, 
#        test_parser,
#        test_lexer_with_file,
#        test_parser_with_file,
        #test_parse_tree,
        #test_traverse_categories,
        #test_get_package,
        test_get_package_classes,
    )
    for t in test_suite:
        try:
            print '--------------- running:', t.__name__
            t()
        except:
            error, det, tb = sys.exc_info()
            print error, det
            traceback.print_tb(tb)

if __name__ == '__main__':
    test()
            
