#
# This module provides a COM-based interface for RationalRose
#

import sys, traceback
import win32com.client as comclient
import pywintypes
import os, os.path
import time


from debug import DEBUG, TRACE, WARNING, ERROR


#
# a function to iterate through RROSE collections
# (wich appear not to be iterable by themselves)
#
def iterate_collection(c):
 	num = c.Count
        #print '*** num = ', num 
 	for idx in range(1, num+1):
            yield c.getAt(idx)



#
# Get a specified set of element properties related to a specific tool
# (this is used to get attribute extra properties in models)
# It returns a dictionary with keys equal to property names and 
# values equal to property values
#
def get_element_properties( element, tool_name, property_names ):
	res = dict()
	for n in property_names:
		try:
			res[n] = element.GetPropertyValue( tool_name, n )
		except  pywintypes.com_error, e:
			WARNING( e ) 
			pass
	return res


#
# Set one or more properties related to a specific tool 
#
def set_element_properties( element, tool_name, new_values_dict ):
	for name, new_value in new_values_dict.items():
		defprop = element.FindDefaultProperty( tool_name, name )
		if element.IsDefaultProperty(tool_name, name):
			is_created = element.CreateProperty(tool_name, defprop.Name, new_value, defprop.Type )
		else:
			prop = element.FindProperty( tool_name, name )
			if prop: prop.Value = new_value
			else: WARNING( ' property %s not found' % name  )
			




#
# Get class data from a class obbject
# Class data are returned as a tuple ( name, stereotype, attrlist, methlist )
# where attrlist is a list of attribute data in the form ( name, type, documentation)
# and methlist is a list of method data in the form ( name, ret_type, paramlist, documentation )
# and param_list is again a list of parameter data in the form ( name, type, init_value, documentation)
# NOTE : if the 'attribute_extra_properties' is not None, it mulst be of type ( tool_name, attribute_names )
# These to elements are used to call get_element_properies and the result, if any, is appended to the 
# atttribute description, which becomes (name, type, documentation, extra_properties).
#
		
def get_class_elements(c, attribute_extra_properties = None ):
	name = c.Name
	stereotype = c.Stereotype
	attributes = []; methods = []
	for a in iterate_collection(c.Attributes):
            a_des = (a.Name, a.Type, a.Documentation )
	    if attribute_extra_properties != None:
		    tool_name, prop_list = attribute_extra_properties
		    a_des += ( get_element_properties( a, tool_name, prop_list ), )
	    attributes.append(a_des)
	for m in iterate_collection(c.Operations):
            plist = []
	    for p in iterate_collection(m.Parameters):
		    p_des = ( p.Name, p.Type, p.InitValue, p.Documentation)
		    plist.append(p_des)
            m_des = (m.Name, m.ReturnType, plist, m.Documentation )
	    methods.append(m_des)
	return name, stereotype, attributes, methods
		




class RRoseInterfaceError(Exception):
    def __init__(self, where, what ):
        Exception.__init__( self, "%s : %s " % (where, what) )


def build_category_path(c, root_category = None):
	parents = [c.Name,]
	while c.ParentCategory != None:
		c =c.ParentCategory
		if c.Name == root_category : break
		parents.append( c.Name )
	parents.reverse()
	return parents


class RRoseInterface(object):
    RROSE_OBJNAME = "Rose.Application"
    def __init__(self):
        "Create the RRose COM client"
        self.client = comclient.Dispatch(self.RROSE_OBJNAME) 
        self.current_model = self.client.CurrentModel
        self.current_category = None

    def exit(self):
        "Quit the client (but does not close RRose)"
        self.client.exit()
        
    def load_model( self, model_file ):
        "Load a model"
        self.client.OpenRoseModel( model_file, False )
        self.current_model = self.client.CurrentModel


    def save_current_model( self, fname=None ):
        assert self.current_model
	if fname:
		res = self.current_model.SaveAs(fname)
	else:
		res = self.current_model.Save
        assert res, "Failed to save the current model"

    def set_current_category( self, *category_path  ):
        "Set the current package. Arguments give the package hierarchy, including package name"
        parent = self.current_model
        cnames = list(category_path)
        current_category = None
        while cnames:
            cname = cnames.pop(0)
            #DEBUG( 'cname', cname )
            all_categories = parent.GetAllCategories
	    current_category = None
            for c in iterate_collection(all_categories ):
                #DEBUG(  c.Name, '<>' , cname, '=>', cname == c.Name )
                if c.Name == cname:
                    current_category = c
                    break
            if current_category:
                parent = current_category
            else:
                raise RRoseInterfaceError("RRoseInterface.set_current_category", 
                                           "Category '%s' not found" % str(cname) )
        self.current_category = current_category

    def iterate_categories(self, *root_category_path ):
        """ Iterates in all sub-categories of current category, including current  """
	if root_category_path:    
		self.set_current_category( *root_category_path )
		TRACE( "iterate_categories - current category ", self.current_category.Name )
		cat_collection =  self.current_category.GetAllCategories
		TRACE( "iterate_categories - cat collection count", self.current_category.GetAllCategories.Count  )
	else:
		cat_collection =  self.current_model.GetAllCategories

	for c in iterate_collection( cat_collection ):
		TRACE( "iterate_categories - Now in category", c.Name )
		cat_path = build_category_path(c)
		self.set_current_category( *cat_path )
		yield c, cat_path



		
    def iterate_on_classes( self, root_category_path, stereotypes=None ):
	    """Iterate on all the classes in the current category and in its sub-categories.
               If a stereotype list is passed, only yields the classes with those stereotypes
	    """
	    self.set_current_category( *root_category_path )
	    for c in self.get_current_category_classes(stereotypes=stereotypes):
		    TRACE( "iterate_on_classes:", 'Processing class:', c.Name, "<<"+c.Stereotype+">>" )
		    yield c
	    self.set_current_category( *root_category_path )
	    for cat, cat_path  in self.iterate_categories( *root_category_path ):  
		TRACE( "iterate_on_classes", 'Now in category:', cat.Name )
		for c in self.get_current_category_classes(stereotypes=stereotypes):
			TRACE( "iterate_on_classes:", 'Processing class:', c.Name, "<<"+c.Stereotype+">>" )
			yield c



    def get_current_category_classes( self, stereotypes = None  ):
        """ Returns the list of classes contained in current category   """
	class_collection = self.current_category.Classes
	return [ c for c in iterate_collection(class_collection) 
		 if ( stereotypes == None) or ( c.Stereotype in  stereotypes ) ]


    def get_current_path(self, root_category = None):
       """Returns the path of current category"""
       if self.current_category:
	  return  build_category_path(self.current_category, root_category)
       else:
          return ""

    def add_category( self, category_name, stereotype=None ):
        """ Add a category   """
        parent = self.current_category
        assert parent
	try:
		p = parent.AddCategory( category_name )
		if stereotype: p.Stereotype = stereotype
		if not p: raise RRoseInterfaceError( "RRoseInterface.add_category",
                                             "Could not create category '%s'" % category_name )
	except pywintypes.com_error, e:
		raise RRoseInterfaceError( "RRoseInterface.add_category",
                                             "Could not create category '%s : %s' " % (category_name, str(e) ) )
        self.current_category = p


    def delete_category( self, category_name ):
        """ Remove a category   """
	parent = self.current_category
	assert parent

	for c in iterate_collection( parent.Categories ):
		if ( c.Name == category_name):
			try :
				parent.DeleteCategory(c)
			except pywintypes.com_error, e:
				raise RRoseInterfaceError( 
					"RRoseInterface.delete_category",
					"Could not delete category '%s : %s' " % (category_name, str(e) ) )

            
    def add_class( self, name, stereotype=None, docum='', attributes=(), methods =() ):
        """
        Add a class to the current category.
        The attributes parameter shall be a  list of elements ( name, type, default_value, documentation )
        The methods parameter shall be a list of elements (name, return_type, param_list, documentation)
        where param_list shall be a list of elements (pname, ptype, pdefault_value, pdocumentation)
        """
        current = self.current_category
        assert current
        cls = current.AddClass( name )
	assert cls, "RRoseInterface.add_class: Failed to create class"+name
        cls.Documentation = docum
        if stereotype: cls.Stereotype = stereotype
        for att in attributes:
            aname, atype, adefault, adoc = att
            attr = cls.AddAttribute( aname, atype, adefault ) 
            assert attr
            attr.Documentation = adoc
        for m in methods:
            mname, ret_type, param_list, mdoc = m
            method = cls.AddOperation( mname, ret_type )
            assert method
            method.Documentation = mdoc
            for idx, p in enumerate(param_list):
                pname, ptype, pdefault, pdoc = p
                par = method.AddParameter( pname, ptype, pdefault, idx )
                assert par
                par.Documentation = pdoc
	return cls


    #
    # Delete a class from current category
    #
    def delete_class( self, class_info ):
        current = self.current_category
        assert current
        current.DeleteClass( class_info )	    


    # Copy in the current category an existing class (from another category). 
    # The 'class_info' parameter is the COM object representing the class to be copied
    def copy_class( self, class_info ):
	    list_attr = [ (a.Name, a.Type.split("::")[-1].strip(), a.InitValue, a.Documentation) 
			  for a in iterate_collection( class_info.Attributes ) ]  
	    list_meth = []
	    for m in iterate_collection(class_info.Operations):
		    param_list = [ (p.name, p.Type, p.InitValue, p.Documentation )
				   for p in iterate_collection(m.Parameters) ]
		    list_meth.append( (m.Name, m.ReturnType.split("::")[-1].strip(), param_list, m.Documentation ) )
	    return self.add_class( class_info.Name, class_info.Stereotype, class_info.Documentation, 
				       list_attr, list_meth )



#
# --- test code after this line
#
import sys, os, os.path


#
# Root of python utils. Required to find all paths inside
#
PYUT_ROOT = os.environ.get("PYTHON_UTILS_ROOT", os.path.realpath(os.path.join("..", ".." )) )

def test():
    TRACE.active = True
    rr_if = RRoseInterface()
    mname = rr_if.current_model.Name
    TRACE( 'Current model is: %s ' % mname )
    rr_if.load_model( os.path.join( PYUT_ROOT, "test", "prova.mdl" ) )
    mname = rr_if.current_model.Name
    TRACE( 'Current model is now: %s ' % mname )

 #   rr_if.set_current_category("Component View", "Implementation Model") 
    #print 'Current category is now: %s ' % rr_if.current_category.Name

    rr_if.set_current_category( "Use Case View", "Capabilities" )
    TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )

    rr_if.set_current_category( "Logical View", "Analysis Model", "Logging" )
    TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )

    # FIXME : for some reason, this category is not listed by GetAllCategories
    #rr_if.set_current_category("Component View") 		
    #TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )
    
    rr_if.set_current_category( "Logical View" )
    TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )

    try:
	    rr_if.delete_category("Prova")
	    TRACE( "Deleted old category Prova" )
    except:
	    WARNING( "Category Prova not deleted - maybe not existing?" )
	    err, det, tb = sys.exc_info()
	    ERROR( err,det )

    rr_if.add_category( "Prova" )
    TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )

    rr_if.add_class("Attore1", stereotype='Actor', docum = 'Esempio di attore.' ) 
    rr_if.add_class("Boundary1", stereotype='boundary', docum = 'Esempio di boundary class.',
                    attributes = ( ('uno', 'integer', '12', 'attributo uno'),
                                   ('due', 'boolean', 'False', 'attributo due'),
                                   ),
                    methods = ( ( 'metodo1', 'boolean', (), 'Metodo senza parametri' ),
                                ( 'metodo2', 
				  'boolean', 
				  ( ('a', 'int', '-10', 'Parameto a'), ), 
				  'Metodo con parametri' )
                                )
                    ) 

				      
    TRACE( "*** ITERATE CATEGOTIES without a root" )
    for c, cpath in rr_if.iterate_categories():
	    TRACE( 'Iterating categories:', cpath, c.Name  )
	    TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )
    TRACE( "*** ITERATE CATEGOTIES with a given root" )
    for c, cpath in rr_if.iterate_categories( "Logical View" ):
	    TRACE( 'Iterating categories:', cpath, c.Name  )
	    TRACE( 'Current category is now: %s ' % rr_if.current_category.Name )


    rr_if.set_current_category( "Prova" )
    classes = rr_if.get_current_category_classes(stereotypes=('boundary'))
    if classes: 
	    r_class = classes [0]
	    rr_if.set_current_category( "Logical View" )
	    TRACE( "Copying the class ", r_class.Name, "in Logical View" )
	    rr_if.copy_class( r_class )
	    rr_if.set_current_category( "Prova" )
	    TRACE( "deleting the class", r_class.Name, "from Prova" )
	    rr_if.delete_class( r_class )

    else:
	 TRACE( "Cancel and recreate class test failed: not class found" )

    TRACE( "Per finire, la lista di tutte le classi nel modello sotto Logical View" )
    for c in rr_if.iterate_on_classes( ("Logical View",) ):
	    print "\t", c.Name, c.Stereotype


   
    rr_if.save_current_model()
    rr_if.exit()




def test_write_properties():	
	rrose_if = RRoseInterface()
	rrose_if.load_model(os.path.join( PYUT_ROOT, "test", "model_with_prop.mdl" ))
	rrose_if.set_current_category( 'Design Model', 'IDD' )
	val=0
	for c in rrose_if.get_current_category_classes():
		for a in iterate_collection(c.Attributes):
			set_element_properties( a, "Attribute Definition", dict([('Size (Bytes)', val), ('Unit', 'Rad')] ))
			val += 1


if __name__ == '__main__':
    #test_write_properties()
    #test_read_extra_details()				    
    test()
			    
