# -*- coding: iso-8859-1 -*-

import sys, traceback
import types
import re

import ptl_parser


def indent_lines(lines, prefix='\t'):
    return [ prefix + l for l in lines ]

#
# Questa classe rappresenta le informazioni estratte dal modello RRose relative
# ad uno specifico elemento. Un elemento pu essere di diverso tipo: le sottoclassi
# di questa classe specializzano ElementInfo per diversi tipii di elementi
#
class ElementInfo( ptl_parser.PTLObject ):
    def __init__(self, pt ):
        ptl_parser.PTLObject.__init__(self, pt)
        if not self.otype: self.otype = ''
        if not self.oname : self.oname = ''
        self.documentation =  self.elements.get('documentation',  "")
        attr_list = self.elements.get("attributes", [["Attributes"],[]] )
        # print "***ElementInfo.__init__ ***", attr_list
        self.tool_data = dict()

        if len(attr_list) > 1 :
            # read element additional tool data, if any 
            for o in attr_list[1]:
                el = ToolDataInfo(o)
                key = self.data_item_key (el.elements.get("tool",""), el.elements.get("name","") )
                #print "storing tool data item", key
                self.tool_data[key] = el

    def data_item_key( self, tool_name, item_name ):
        return tool_name + "." + item_name
            
    def get_tool_data_item(self, tool_name, item_name):
        try:
            key = self.data_item_key(tool_name, item_name)
            value = self.tool_data[key].item_value
            if (type(value) in (types.ListType, types.TupleType)) and (value[0] == 'value'):
                value = value[-1]
            #print "***",  key, self.tool_data[key]
            return value.strip()
        except KeyError:
            return ""
        
    def tool_data_lines(self):
        res = []
        for t in self.tool_data.values():
            res += indent_lines(t.to_lines() )
        return res
    

    def to_lines(self):
        res = [self.otype + ' ' + self.oname, self.documentation ]
        for el in self.tool_data.values():
            res += indent_lines(el.to_lines())     
        return res    

    def __repr__(self):
        return '\n'.join( self.to_lines() + self.tool_data_lines() )


#
# Questa classe specializza la classe ElementInfo per un generico set di dati di un 'tool'
# 
#
class ToolDataInfo( ElementInfo ):
    def __init__(self, pt):
        ElementInfo.__init__(self, pt)
        self.tool = self.elements.get("tool", "")
        self.item_name = self.elements.get("name", "")
        self.item_value = self.elements.get("value", "")
        self.quit = self.elements.get("quit", "")

    def to_lines(self):
        return [ self.tool + "." + self.item_name + "=" + str(self.item_value)] 


#
# Questa funzione fa il parsing di un attributo 'multiplicity' in formato Val(nome_attributo)
# Se ci riesce, restituisce 'nome_attributo'
# Se non ci riesce, restituisce la stringa di input
#
multiplicity_re = re.compile("Val\((\w+)\)") 
def parse_multiplicity(m_str):
    name = multiplicity_re.findall(m_str)
    if name: return name[0]
    else: return m_str
    




#
# Questa classe specializza ElementInfo per estrarre da uno specifico data tool alcume
# informazioni relative agli attributi
#
class AttributeInfo( ElementInfo ):
    TOOLNAME = "Level D Definition"
    MULTIPLICITY_ATTRNAME = "Multiplicity"
    SIZE_ATTRNAME = "Size (Bytes)"
    MIN_ATTRNAME = "Min Value"
    MAX_ATTRNAME = "Max Value"
    RESOLUTION_ATTRNAME = "E_LSB"
    UNIT_ATTRNAME = "Unit"
    INITIAL_VALUE_ATTRNAME = "initv"
    LEVEL_D_ATTRS = ( SIZE_ATTRNAME, MULTIPLICITY_ATTRNAME, MIN_ATTRNAME, MAX_ATTRNAME, RESOLUTION_ATTRNAME, UNIT_ATTRNAME )

    def __init__(self, pt):
        ElementInfo.__init__(self, pt)
        self.type = self.elements.get("type", "type not found for "+self.oname )
        #if self.type == "type?":
        #    print "Warning: type not found for attribute", self.oname
        # attr_list = self.elements.get("attributes", [["Attributes"],[]] )
        self.initial_values = self.elements.get( self.INITIAL_VALUE_ATTRNAME, "" )

    def get_level_d(self):
        return dict( (n, self.get_tool_data_item( self.TOOLNAME, n)) for n in  self.LEVEL_D_ATTRS )
    

    def get_multiplicity(self):
        value = self.get_tool_data_item( self.TOOLNAME, self.MULTIPLICITY_ATTRNAME)
        value = parse_multiplicity(value)
        try:
            return int(value)
        except (TypeError, ValueError):
            return value

    def get_size(self ):
        size_str = self.get_tool_data_item(self.TOOLNAME,self.SIZE_ATTRNAME )
        if not size_str: size = 4 # default
        else:
            try:
                size = int(size_str)
            except ValueError:
                try:
                    size = float(size_str)
                except ValueError:
                    size = size_str
        return size

    def to_lines(self):
        res = [self.type + " " + self.oname + "\t" + self.documentation ]
        return res




class OperationParameterInfo( ElementInfo ):
    TOOLNAME = "Additional Parameter Data"
    INOUT_PROPERTY_NAME = "in/out"
    OPTION_PROPERTY_NAME = "Option"
    
    def __init__(self, pt):
        ElementInfo.__init__(self, pt)
        self.type = self.elements.get("type", "type?")
        self.inout = self.get_tool_data_item( self.TOOLNAME, self.INOUT_PROPERTY_NAME)
        self.option = self.get_tool_data_item( self.TOOLNAME, self.OPTION_PROPERTY_NAME)
        
    def to_lines(self):
        return [self.type + ' ' + self.oname + '\t' + self.inout + self.documentation + self.option ]  
    

class OperationInfo(ElementInfo):
    def __init__(self, pt ):
        ElementInfo.__init__(self, pt)
        self.return_type = self.elements.get('r', 'type?')
        paramlist = self.elements.get('parameters', ['Parameters', []] )
        self.parameters = [ OperationParameterInfo(o) for o in paramlist[1] ]
 
    def to_lines(self):
        res = ElementInfo.to_lines(self)
        for p in self.parameters: res += indent_lines( p.to_lines() )
        return res

class ClassInfo(ElementInfo):
    CLS_TOOLNAME = "Additional Class Data"
    MSG_TOOLNAME = "Additional Message Data"
    IRS_REFERENCE_ATTR_NAME = "IRS Requirement"
    IDENTIFIER_ATTR_NAME ="Identifier"
    OCCURRENCE_ATTR_NAME = "Occurrence"
    EXPECTED_ANSWER_ATTR_NAME = "Expected answer"
        
    def __init__(self, pt ):
        ElementInfo.__init__(self, pt)
        oplist = self.elements.get( 'operations', ['Operations', []] )
        self.operations = [OperationInfo(o) for o in oplist[1] ]
        attrlist = self.elements.get('class_attributes', ['Attributes', []])
        self.attributes = [AttributeInfo(o) for o in attrlist[1] ]
        self.stereotype = self.elements.get('stereotype', "")
        superlist =  self.elements.get('superclasses', ['inheritance_relationship_list', []])
        self.superclasses = [ ElementInfo(sc).elements.get("supplier").split("::")[-1] for sc in superlist[1] ]
        self.irs_ref = self.get_tool_data_item( self.MSG_TOOLNAME, self.IRS_REFERENCE_ATTR_NAME )
        self.identifier = self.get_tool_data_item( self.CLS_TOOLNAME, self.IDENTIFIER_ATTR_NAME )
        self.occurrence = self.get_tool_data_item( self.CLS_TOOLNAME, self.OCCURRENCE_ATTR_NAME )
        self.expected_answer = self.get_tool_data_item( self.CLS_TOOLNAME, self.EXPECTED_ANSWER_ATTR_NAME )

    def to_lines(self):
        res = ElementInfo.to_lines(self)
        res += ["IRS Requirement:"+ self.irs_ref, "Identifier:" + self.identifier, 
                "Occurrence:" + self.occurrence, "Expected answer:" + self.expected_answer ]
        #res += ["Superclasses:" + ' '.join(self.superclasses)]
        res += ['\tOperations:']
        for op in self.operations: res += indent_lines(op.to_lines (), '\t\t')
        res += ['\tAttributes:']
        for att in self.attributes: res += indent_lines(att.to_lines(), '\t\t')
        return res

    def __repr__(self):
        return '\n'.join(self.to_lines())


#
# Class to represent relevand data about diagrams
#
class DiagramInfo(ElementInfo):
    def __init__(self, pt, nested_level=1 ):
        ElementInfo.__init__(self, pt )
        self.diagram_type = pt[1]
        self.title = self.elements.get("title", self.oname)
 

# This class represents the category infornmation extraced from RRose File
# Each category  may contain other categories as well as class info 
class CategoryInfo(ElementInfo):
    def __init__(self, pt, nested_level=1 ):
        ElementInfo.__init__(self, pt )
        self.stereotype = self.elements.get('stereotype', "")
        self.categories = [] 
        self.classes = []
        self.nested_level = nested_level 

        # trova i class diagram
        diag_list = self.elements.get("logical_presentations", []) 
        if diag_list and len(diag_list) > 1:            
            self.diagrams = [ DiagramInfo(d) for d in diag_list[1] if d[1] == "ClassDiagram" ]
            #print "*** Package %s - found %d diagrams: %s" % ( self.oname, len(self.diagrams), ",".join( d.oname for d in self.diagrams ))
        else:
            self.diagrams = []
            
        #trova gli activity diagrams
        act_diag = self.elements.get("statemachine", None)
        if act_diag:
            #print act_diag
            self.diagrams.append( DiagramInfo(act_diag) )
            
        
        # trova le classi e le sotto-categorie
        contents = self.elements.get("logical_models", None)
        if contents and len(contents) > 1: 
            contents = contents[1]
        else: 
            contents = []
        for e in contents:
            if e[1] == 'Class_Category':
                self.categories.append( CategoryInfo(e, nested_level+1 ) )
            elif e[1] == 'Class':
                self.classes.append( ClassInfo(e) )
        
    
    # Generatore che percorre l'albero delle categorie depth-first, restituendo di volta in volta una
    # categoria o una classe ( o solo classi o solo categorie )
    def walk(self, classes_only = False, categories_only = False):
        to_do = self.classes + self.categories
        while to_do:
            o = to_do.pop()
            if ( classes_only == True ) and ( type(o) == ClassInfo ):
                yield o
            elif ( categories_only == True ) and ( type(o) == CategoryInfo ):
                yield o
            elif ( classes_only == False ) and ( categories_only == False ):
                yield o
                
            if type(o) == CategoryInfo:
                to_do.extend(o.classes)
                to_do.extend(o.categories) 
                                

    def to_lines(self):
        res = ElementInfo.to_lines(self)
        res.append( "\tClasses:" )
        for c in self.classes: res += indent_lines( c.to_lines(), "\t\t")
        res.append( "\tCategories:" )
        for c in self.categories: res += indent_lines( c.to_lines(), "\t\t")
        return res
        
    def __repr__(self):
        return '\n'.join(self.to_lines())
        

#
# Funzione-generatore che restituisce le class info estratte da un file-modello RRose
#
def read_class_info ( model_file, category_name ):
    pt = ptl_parser.ParseTree()
    pt.parse_file( model_file )
    for e in pt.get_package_elements( category_name, ('object', 'Class') ):
        yield ClassInfo(e)

#
# Funzione-generatore che restituisce le categorie estratte da un documento RationalRose
# con specifici nomi. 
#
def read_category_info( model_file, *category_names ):
    pt = ptl_parser.ParseTree()
    pt.parse_file( model_file )
    categories = dict()
    current=category ="Undefined"
    for e in pt.traverse( ):
        if e and len(e) > 2 and  e[1]  == 'Class_Category' and e[2] in category_names:
            #print "Found category", e[:3]
            yield CategoryInfo(e)
            
        




#
# ----------------- test code after this line 
#
import os, os.path
PYUT_ROOT = os.environ.get("PYTHON_UTILS_ROOT",
                           os.path.realpath(os.path.join("..", "..")))

TEST_MODEL = os.path.join( PYUT_ROOT, "test", "uwpsmf_sdd_idd_modell_7.7.mdl" )

def test_read_attribute_info():
    for c in read_class_info( TEST_MODEL, "BMS Messages" ):
        print '-'*80
        print "Attributes of class:", c.oname
        for a in c.attributes:
            print a
            print '\t\tsize =', a.get_tool_data_item("Attribute Definition", "Size (Bytes)")

def test_read_class_info():
    for c in read_class_info( TEST_MODEL, "Utilities" ):
        print c
    for c in read_class_info( TEST_MODEL, "BMS_Messages " ):
        print c


def test_read_category_info():
    for cat in read_category_info( TEST_MODEL, "Concept of Execution", "Application"  ):
        for ci in cat.walk():
            if type(ci) == CategoryInfo:
                print "Package:", ci.oname
                for d in ci.diagrams:
                    print d.diagram_type, ":", d.title
            else:
                print "Class:", ci.oname
 
if __name__ == '__main__':
    #test_read_attribute_info()
    #test_read_class_info()
    test_read_category_info()
            
