import sys, traceback
import re

from win32com.client import constants, Dispatch
import pythoncom, pywintypes

def get_table_contents(table_object):
  res = []
  for r in table_object.Rows:
    res.append([])
    for c in r.Cells:
      res[-1].append( c.Range.Text )
  return res  

class Fontproperties(object):
    def __init__(self, name=None, size=None, bold=True, italic=True ):
        self.fname = name
        self.size = size
        self.bold = bold
        self.italic = italic
        
class TableProperties(object):
    def __init__(self, header_font, header_color, column_sizes, border ):
        self.header_font = header_font
        self.header_color = header_color
        self.border = border
        self.column_sizes = column_sizes[:]
        


class WordInterface(object):
  """
  Some convenience methods for Word documents accessed
  through COM.
  Taken from http://www.bigbold.com/snippets/posts/show/2037
  and changed/extended
  """
  
  def __init__(self, visible=False):
    self.app = Dispatch("Word.Application")
    self.set_visible(visible)
    self.table_properties = TableProperties(header_font=None, header_color=None, column_sizes=[], border=True)
    
  def set_visible(self, flag):
      self.app.Visible = flag
  
  def new_document(self,  template=None):
    """
    Create a new Word document. If 'filename' specified,
    use the file as a template.
    """
    if template: doc = self.app.Documents.Add(template)
    else: doc = self.app.Documents.Add()
    return doc
   
  def open_document(self, filename):
    """
    Open an existing Word document for editing.
    """
    try:
      doc = self.app.Documents.Open(filename)
    except pywintypes.com_error:
      raise Exception("Unable to open the file: %s, - %s" % ( filename, str(sys.exc_info()[1] ) ))
    return doc
  
  def save_active(self, filename=None):
    """
    Save the active document. If 'filename' is given,
    do a Save As.
    """
    if filename:
      self.app.ActiveDocument.SaveAs(filename)
    else:
      self.app.ActiveDocument.Save()
  
     
  def close_active(self):
    """
    Close the active document.
    """
    self.app.ActiveDocument.Close()

  
  def quit(self):
    """
    Quit Word.
    """
    return self.app.Quit()


  def insert_paragraph(self, text, stylename=None, bold=False, italic=False, underline=False):
    p = self.app.ActiveDocument.Paragraphs.Add()
    r = p.Range
    if stylename: r.Style = self.app.ActiveDocument.Styles(stylename)
    if bold: r.Font.Bold = 1
    if italic: r.Font.Italic = 1
    if underline: r.Font.Underline = 1
    r.InsertBefore(text)
    return p

  def get_next_section(self):
    for s in self.app.ActiveDocument.Sections:
      yield s

  def get_next_paragraph(self, *stylenames ):
    for p in self.app.ActiveDocument.Paragraphs:
      #print p.Style
      if not stylenames or str(p.Style) in stylenames:
        yield p.Style, p.Range.Text
      #else: print '.'
    return


  def get_next_paragraph_object(self, *stylenames ):
    for p in self.app.ActiveDocument.Paragraphs:
      #print p.Style
      if not stylenames or str(p.Style) in stylenames:
        yield p
      #else: print '.'
    return


  def get_next_table(self):
    res = []
    for t in self.app.ActiveDocument.Tables:
      yield get_table_contents(t)


  def replace_table_element(self, table_idx, row_idx, col_idx, new_text):
    t = self.app.ActiveDocument.Tables[table_idx]
    c = t.Rows[row_idx].Cells[col_idx]
    c.Range.Select()
    self.app.Selection.TypeText(new_text)
    
  def process_table_rows(self, row_processor, tables = None,  ):
    for t_idx, t in enumerate( self.app.ActiveDocument.Tables ):
      if not tables or t_idx in tables:
        for r in t.Rows:
          row_contents = [ c.Range.Text for c in r.Cells ]
          row_changes = row_processor(row_contents)
          for idx, el in row_changes:
            if el != None:
              r.Cells[idx].Range.Select()
              self.app.Selection.TypeText(el)

  def search_replace_rows(self, *rules  ):
    for t in self.app.ActiveDocument.Tables:
      for r in t.Rows:
        row_contents = [ c.Range.Text for c in r.Cells ]
        for pattern, replacement in rules:
          # compare the row elements with the non-null elements of row_pattern
          if len(row_contents) != len(pattern) : continue
          match = True
          for idx, el in enumerate(pattern):
            if el != None:
              if not re.findall( el, row_contents[idx] ): 
                match = False
                break # skip the other comparisons if one fails
          if match: # replace with non-null elements of row_replacement
            for idx, el in enumerate(replacement ):
              if el != None: 
                r.Cells[idx].Range.Select()
                self.app.Selection.TypeText(el)
            break # if one rule matches, don't check the otrher rules

  def replace_field_text(self, old_text, new_text):
    for f in self.app.ActiveDocument.Fields:
      code_txt = f.Code.Text
      if old_text in code_txt:
        new_code_txt = code_txt.replace(old_text, new_text)
        #print code_txt, new_code_txt  
        f.Code.Delete()
        f.Code.InsertBefore(new_code_txt)
        
  def set_table_properties(self, header_font= None, header_color=None, borders = True, column_sizes=[] ):
      self.table_properties = TableProperties(header_font, header_color, column_sizes, borders )

  def insert_vertical_table(self, caption=None, headers=(), rows=(), caption_type='Table', style=None ):
    """
    Create a table with headers as first row (in bold & italic)
    """
    if headers: numcolumns = len(headers)
    elif rows: numcolumns = len(rows[0])
    else: numcolumns = 0
    numrows = len(rows) + 1
    doc = self.app.ActiveDocument
    p = self.app.ActiveDocument.Paragraphs.Add()
    t = self.app.ActiveDocument.Tables.Add( p.Range, numrows, numcolumns)

    if style: t.Style = style
    
    # Enable default borders for the table
    t.Borders.Enable = self.table_properties.border;


    
    # insert headers in first row, in bold&italic if required
    if headers:
        if self.table_properties.header_font:
            hf_italic = 1 if self.table_properties.header_font.italic else 0
            hf_bold = 1 if self.table_properties.header_font.bold else 0
        else:
            hf_italic, hf_bold = 0,0
            
        r = doc.Range(Start=t.Rows(1).Cells(1).Range.Start, End= t.Rows(1).Cells(len(headers)).Range.End)
        r.Font.Bold = hf_bold; r.Font.Italic = hf_italic
        for idx, h in enumerate(headers):
              t.Rows(1).Cells(idx+1).Range.InsertBefore(h)
            
        #
        # Set header shading
        #
        if self.table_properties.header_color:
            color = self.table_properties.header_color
            #t.Rows(1).Shading.Texture = texture # 20 percent
            t.Rows(1).Shading.ForegroundPatternColor = color
    

    #insert contents
    for i, r in enumerate(rows):
      for j, txt in enumerate(r):
          t.Rows(i+2).Cells(j+1).Range.InsertBefore(txt)
    
    # Set column size or fit to contents property
    col_sizes = self.table_properties.column_sizes
    if not col_sizes :
        t.AutoFitBehavior(1); # fit to contents
    else:
        for idx, csize in enumerate(col_sizes):
            if csize != None: 
                c = t.columns[idx]
                c.PreferredWidthType = 2 # percent
                c.PreferredWidth = csize
                
                
    
    #insert caption
    if caption:
      self.insert_caption( caption_type, caption )


  def insert_caption(self, caption_type, caption_text):
    p = self.insert_paragraph("", stylename='Caption')
    p.Range.Select()
    self.app.Selection.InsertCaption(caption_type, " - " + caption_text)
          
  def find_and_replace(self, find_str, replace_str):
    """
    Find all occurances of 'find_str' and replace with 'replace_str'
    in the active document.
    """
    self.app.Selection.HomeKey(Unit=wdStory)
    find = self.app.Selection.Find
    find.Text = find_str
    while self.app.Selection.Find.Execute():
      self.app.Selection.TypeText(Text=replace_str)


#
# ------ Test code after this line
#

def test_1():
    w = WordInterface();
    w.app.Visible=1
    w.new_document()

    w.insert_paragraph( "ciao Moddo :-)",  stylename="Heading 1")
    w.insert_paragraph( "" )
    w.insert_paragraph("hello word :-/")    
    w.insert_paragraph( "This is in bold and Italic", bold=True, italic=True)
    w.insert_paragraph( "This is underlined.", underline = True )
    w.insert_paragraph( "" );

    w.insert_vertical_table( "This is a table with default properties and predefined style",
                             ("First Column", "Second Column", "Third Column"),
                             style = "IDDTableStyle",
                    rows=[("1", "2", "3"),
                          ("11", "22", "33"),
                          ("aa", "bb", "CCccCC")],
                  )
    
    w.set_table_properties(header_font=Fontproperties(name=None, italic=True, bold=True), 
                           #header_color=33, 
                           borders=False, 
                           column_sizes = [0.2, 0.2, 0.6])
    w.insert_vertical_table( "This is a table with specific properties and the same style",
                             ("First Column", "Second Column", "Third Column"),
                    [("1", "2", "3"),
                    ("11", "22", "33"),
                    ("aa", "bb", "CCccCC-----------------------------------")],
                    style = "IDDTableSTyle"
                  )
    
    
    
                    
    #w.save_current()
    

def test_tables_1():
  w = WordInterface()
  w.open_document()
  for t in w.get_next_table(r'../test/prova_word_interface.doc'):
    print t
  w.replace_table_element(table_idx=0, row_idx=1, col_idx=2, new_text='Paperino')
  w.set_visible(True)

def test_tables_2():
    w = WordInterface()
    w.open_document(r'../test/prova_word_interface.doc')
    w.search_replace_rows( ( ( None, None, u'float', None, None, u'tbd', u'tbd' ), # pattern 
                             ( None, None, None,     None, None, u'N/A', u'N/A' ) # replacement
                           ), # rule end
                           ( ( None, None, u'boolean', None,   None,         u'tbd', u'tbd' ), # pattern 
                             ( None, None, None,       u'N/A', u'True/False', u'N/A', u'N/A' ) # replacement
                           ) # rule end
                         )
    w.set_visible(True)


def test_row_processor():
  re_float = re.compile('float')
  re_tbd = re.compile('TBD')

  def row_processor( row_contents ):
    res = []
    if re_float.findall( row_contents[1] ) and re_tbd.findall(row_contents[4]):
      res.append ((4, 'N/A'))
    return res
      
  w = WordInterface()
  w.open_document(r'../Prova1.doc')
  w.process_table_rows( row_processor )
  w.set_visible(True)
  

if __name__ == '__main__':
    test_1()
    test_row_processor()
    test_tables_2()
    print 'Done.'
    
